#!/usr/bin/perl
#----------------------------------------------------------------------------
# $Id: slissom.pl,v 1.4 2000/05/25 00:15:40 yschoe Exp yschoe $
#	slissom.pl
#	- implements SLISSOM 
#	- uses an exact mathematical description of the model
#	  rather than an algorithmic approach.
#----------------------------------------------------------------------------
use Getopt::Long;

$usage = <<'EOF';
usage: slissom.pl [options]

	options: parameters and default values

EOF
# Usage is continued after the following section.

#----------------------------------------
# Section: Parameters: %Params is global
#----------------------------------------
%Params = (
	"verbose+" => 0,
	"lambda_exc_syn=f"  => 0.1,	# exc synapse decay rate
	"lambda_inh_syn=f"  => 0.1,	# inh synapse decay rate
	"lambda_int_syn=f"  => 0.1,	# intra columnar synapse decay rate
	"lambda_loop=f" => 0.5,		# inh loop decay rate
	"gamma_aff=f" 	=> 1.1,		# aff contri
	"gamma_int=f"	=> 1.2,		# int columnar contri
	"gamma_exc=f"	=> 1.3,		# exc contri
	"gamma_inh=f"	=> 1.8,		# inh contri
	"tau=f"		=> 1.0,		# inh loop scaling
	"theta_base=f"		=> 0.15, # base thresh
	"alpha=f"	=> 0.0,		# sigmoid thresh
	"beta=f"	=> 1.0,		# sigmoid ceiling
	"iterations=i"	=> 100,		# simulation duration
	"neurons=i"	=> 3,		# number of neurons
	"input_select=s" => "square_pulse",	# type of input
	"print_isi_flag+" => "0",
	"srand=i" => "12345",		# random seed
	"exc_radius=i" => 0,		# excitatory radius
	"inh_radius=i" => 0,		# excitatory radius
	"random_init_neurons+" => 0,	# initialize neurons randomly? 
	"exc_w=f" => 1.0,		# initial exc weight
	"inh_w=f" => 1.0,		# initial inh weight
	"self_lateral=i" => 1,		# lateral connection to self 1:on 0:off
	"novisible_spike+" => 0,	# do not add 1.0 to visualize spike
	"wgt_profile=s" => "uniform",	# initial weight profile
	"plot_weights+" => 0,		# plot weights
	"normalize+" => 0,		# normalize weights
	"noise=f" => 0.0,		# noise level [0.0-1.0]
	"aff_noise=f" => 0.0,		# input noise level [0.0-1.0]
	"hard_refractory=i" => 0.0,	# hard refractory period length 
	"print_contrib=i" => 0,	# neuron# of stat to print
	"patchy_exc=i" => 0,		# patchy exc connections? # patches 
	"notime" => 0
);

foreach $key (sort(keys(%Params))) {
	($name,$type) = split(/[=:]/,$key,2);
	$usage .= "\t--$name   \t(default = $Params{$key})\n";
}
$usage .= "\n";

#----------------------------------------
# Section: Read parameters from command line
#----------------------------------------
my $option_str = join('","',keys(%Params));
eval("GetOptions(\"$option_str\",help)"); 
if ($opt_help) {
  print STDERR $usage;
  exit(0);
}

#----------------------------------------
# Subsection: put parameters in %Params hash
#----------------------------------------
my $param, $opt_name, $opt_type;
foreach $param (keys(%Params)) {
  ($opt_name,$opt_type) = split(/[=:]/,$param,2);
  $opt_varname = "opt_$opt_name";
  if (!$$opt_varname) {
	$$opt_varname = $Params{"$param"};
  }
  $Params{"$param"} = $$opt_varname;
}

#----------------------------------------
# Subsection: print out parameters
#----------------------------------------
print STDERR "\n----------------------------------------\n" if ($opt_verbose);
$today = `date +"%b,%d %Y"`; chop $today;
print STDERR "\n<h1> $today : </h1> \n" if ($opt_verbose);
print STDERR "\n<pre>\n" if ($opt_verbose);
foreach $param (sort(keys(%Params))) {
	printf STDERR "% 25s",$param ;
	printf STDERR " = $Params{$param}\n" if ($opt_verbose);
}
print STDERR "\n</pre>\n" if ($opt_verbose);
print STDERR "<img src=\"\">\n";
print STDERR "----------------------------------------\n" if ($opt_verbose);

#----------------------------------------
# Section: Set random seed
#----------------------------------------
srand($opt_srand);

#----------------------------------------
# Section: Initialize Neurons
#----------------------------------------

$slissom = Network::new(
	{"x" => $opt_neurons,
	 "y" => 1
	}
);
	 
@n = [];

for ($i=0; $i<$opt_neurons; ++$i) {

  my $eta, $theta;
  if ($opt_random_init_neurons) {
	$eta = -rand(1.0);	
	$theta = rand(1.0);	
  } else {
	$eta = 0.0;
	$theta = 0.0;
  }
  $n[$i] = Neuron::new(
	{"x" 		=> $i, 
	"y" 		=> 0,
	"lambda_exc_syn" 	=> $opt_lambda_exc_syn,
	"lambda_inh_syn"  	=> $opt_lambda_inh_syn,
	"lambda_int_syn"  	=> $opt_lambda_int_syn,
	"lambda_loop" 	=> $opt_lambda_loop,
	"gamma_aff" 	=> $opt_gamma_aff,
	"gamma_int"	=> $opt_gamma_int,
	"gamma_exc"	=> $opt_gamma_exc,
	"gamma_inh"	=> $opt_gamma_inh,
	"tau"		=> $opt_tau,
	"alpha"		=> $opt_alpha,
	"beta"		=> $opt_beta,
	"theta_base"	=> $opt_theta_base,
	"input_select"	=> $opt_input_select,
	"print_isi_flag" => $opt_print_isi_flag,
	"state"		=> [ "t", "new_s_exc", "new_s_inh", "new_s_int",
			"new_eta", "net_aff", "net_int", "net_exc", "net_inh",
			"new_theta", "spike", "v", "r"],
	"eta"		=> $eta,
	"theta"		=> $theta,
	"novisible_spike" => $opt_novisible_spike,
	"noise" => $opt_noise,
	"aff_noise" => $opt_aff_noise,
	"hard_refractory" => $opt_hard_refractory,
	"network"	=> \@n
	}
  );
  $n[$i]->show_obj();
}

#----------------------------------------
# Section: Initialize Connections
#----------------------------------------

# $n[0]->connect($n[1],"exc",$opt_exc_w);
# for ($i=1; $i<$opt_neurons-1; ++$i) {
# 	$n[$i]->connect($n[$i-1],"exc",$opt_exc_w);
# 	$n[$i]->connect($n[$i+1],"exc",$opt_exc_w);
# }
# $n[$opt_neurons-1]->connect($n[$opt_neurons-2],"exc",$opt_exc_w);
for ($i=0; $i<$opt_neurons; ++$i) {

  my $s,$e;

  # Excitatory connections
  if ($opt_exc_radius>0) {
    $s = $i-$opt_exc_radius;
    $e = $i+$opt_exc_radius;
    $s = (($s)<0)?0:$s;
    $e = (($e)>=$opt_neurons)?($opt_neurons-1):$e;

    if ($opt_patchy_exc>0) {
	my $my_patch = int(($opt_patchy_exc*($i-1)+$opt_patchy_exc)
				/ ($opt_neurons*1.0));
	for ($k=0; $k<$opt_patchy_exc; ++$k) {
		# - find out what patch this neuron is on
		# - connect to alternating patches only
		if (abs($k-$my_patch)%2==0) { 
		   my $this_patch_s = int($k*$opt_neurons
				    / ($opt_patchy_exc*1.0));
		   my $this_patch_e = int(($k+1)*$opt_neurons
				    / ($opt_patchy_exc*1.0)-1);
		   $this_patch_s=($this_patch_s<$s)?$s:$this_patch_s;
		   $this_patch_e=($this_patch_e>$e)?$e:$this_patch_e;
    		   $n[$i]->connect_within_range($this_patch_s,$this_patch_e,
			"exc",$opt_exc_w,
			$opt_self_lateral,
			$opt_wgt_profile,$opt_exc_radius,
			$i);
			# int(($this_patch_e-$this_patch_s)/2.0));
		}
	}
    } else {
    	$n[$i]->connect_within_range($s,$e,"exc",$opt_exc_w,$opt_self_lateral,
		$opt_wgt_profile,$opt_exc_radius,$i);
    }
    $n[$i]->normalize_weights("exc") if ($opt_normalize);
  }

  # Inhibitory connections
  if ($opt_inh_radius>0) {
    $s = $i-$opt_inh_radius;
    $e = $i+$opt_inh_radius;
    $s = (($s)<0)?0:$s;
    $e = (($e)>=$opt_neurons)?($opt_neurons-1):$e;
    $n[$i]->connect_within_range($s,$e,"inh",$opt_inh_w,$opt_self_lateral,
		$opt_wgt_profile,$opt_inh_radius,$i);
    $n[$i]->normalize_weights("inh") if ($opt_normalize);
  }

}

if ($opt_plot_weights) {

  for ($i=0; $i<$opt_neurons; $i += ($opt_neurons/10)) {
	$n[$i]->plot_weights("exc",0);
  }
  exit;
}

#----------------------------------------
# Section: Run simulation
#----------------------------------------
for ($i=0; $i<$opt_iterations; ++$i) {

  # print STDERR "running $i\n" if ($opt_verbose);
  for ($k=0; $k<$opt_neurons; ++$k) {
	$n[$k]->calc_next_state($i);
	$n[$k]->show_state("neuron$k");
	$n[$k]->advance_state();
  }
  print "$i " if (!$opt_notime);
  for ($k=0; $k<$opt_neurons; ++$k) {
	print $n[$k]->{"v"}." ";
  }
  print " allv\n";

  for ($k=0; $k<$opt_neurons; ++$k) {
	print $n[$k]->{"spike"}." ";
  }
  print "spikes\n";

  print "$i"." ";
  print $n[$opt_print_contrib]->{"net_aff"}." ";
  print $n[$opt_print_contrib]->{"net_exc"}." ";
  print $n[$opt_print_contrib]->{"net_inh"}." contribs\n";

}

# Print what each column means : exclude the first column when counting --
# first columns are usually "neuron<n>" where <n> is a number.
$n[0]->show_state_legends() if ($opt_verbose);

# Print ISI histogram
if ($opt_print_isi_flag) {
	$n[0]->show_isi_histogram("isi0");
}

#----------------------------------------------------------------------------
#
# 	Subroutines
#
#----------------------------------------------------------------------------

#----------------------------------------------------------------------------
# Init vector of size '$size'
#----------------------------------------------------------------------------
sub init_vector {

  my ($size) = @_;
  my $i;
  my @vector;

  for ($i=0; $i<$size; ++$i) {
	$vector[$i] = $0;
  }

  return @vector;
}

#----------------------------------------------------------------------------
#
# Neuron package definition
#
#----------------------------------------------------------------------------
package Neuron;

sub new {
	my $spec = shift;

 	my $neuron = {}; # reference to an anonymous hash

	# Copy attributes
	foreach $key (keys(%$spec)) {
		$neuron->{"$key"} = $spec->{"$key"};
	}
	
	# init states
	$neuron->{"t"} 	   = 0;		# global time
	$neuron->{"s_exc"} = 0.0;
	$neuron->{"s_inh"} = 0.0;
	$neuron->{"s_int"} = 0.0;
	$neuron->{"r"} = 0.0;		# refractory
	$neuron->{"eta"} = 0.0;
	$neuron->{"delta"} = 0.0;
	$neuron->{"h_ref"} = 0.0;

	$neuron->{"new_s_exc"} = 0.0;
	$neuron->{"new_s_inh"} = 0.0;
	$neuron->{"new_s_int"} = 0.0;
	$neuron->{"new_r"} = 0.0;		# refractory
	$neuron->{"new_eta"} = 0.0;
	$neuron->{"new_delta"} = 0.0;

	$neuron->{"exc_w"} = {}; 
	$neuron->{"inh_w"} = {};
	$neuron->{"aff_w"} = {};
	$neuron->{"int_w"} = {};

	$neuron->{"exc_decay"} = exp(-$neuron->{"lambda_exc_syn"});
	$neuron->{"inh_decay"} = exp(-$neuron->{"lambda_inh_syn"});
	$neuron->{"int_decay"} = exp(-$neuron->{"lambda_int_syn"});
	$neuron->{"loop_decay"} = exp(-$neuron->{"lambda_loop"});

	$neuron->{"spike"} = 0;

	$neuron->{"prev_spike_t"} = 0;
	$neuron->{"isi_histogram"} = {};

	bless $neuron, 'Neuron';
	return $neuron; # return reference to anonymous hash

}

#----------------------------------------
#	show_stat
#----------------------------------------
sub show_obj {

	my $neuron = shift;

	print "Neuron Stat [$neuron->{x}] ------------------------------\n";
	foreach $key (sort(keys(%$neuron))) {
		print "$key    \t\t= $neuron->{$key}\n";
	}
	print "\n";
		
}

# Print selected state variables 
sub show_state {

	my $neuron = shift;
	my $tag = shift;

	my $sref = $neuron->{"state"};
	foreach $key (@$sref) {
		print "$neuron->{$key} " if ($key ne "new_theta");
		print -$neuron->{$key}." " if ($key eq "new_theta");
	}
	print "$tag\n"; 
		
}

# Print legends to the selected state variables
sub show_state_legends {

	my $n = shift;
	my $i=1;
	my $aref = $n->{"state"};
	my $key;
	foreach $key (@$aref) {
	  print STDERR "$i:$key\n";
	  $i++;
	}
	print STDERR "\n";
}



#----------------------------------------
#	Calc state
#----------------------------------------
sub calc_next_state {

	my $n = shift;

	# Get global time
	$n->{"t"} = shift;

	# Spike
	$n->{"spike"} = $n->Heav($n->{"eta"} - $n->{"theta"});

	# Accumulate ISI histogram
	if ($n->{"print_isi_flag"} && $n->{"spike"}==1.0) {
		my $isi_hist = $n->{"isi_histogram"};
		my $isi = $n->{"t"}-$n->{"prev_spike_t"};

		if (exists($isi_hist->{"$isi"})) {
			$isi_hist->{"$isi"} = $isi_hist->{"$isi"} + 1;
		} else {
			$isi_hist->{"$isi"}=1;
		}
		$n->{"prev_spike_t"} = $n->{"t"};
	}

	# Leaked output
	$n->{"new_s_exc"} = $n->{"s_exc"}*$n->{"exc_decay"} + $n->{"spike"};
	$n->{"new_s_inh"} = $n->{"s_inh"}*$n->{"inh_decay"} + $n->{"spike"};
	$n->{"new_s_int"} = $n->{"s_int"}*$n->{"int_decay"} + $n->{"spike"};

	# Contributions
	$n->{"net_aff"} = $n->aff_contri();
	$n->{"net_int"} = $n->int_contri();
	$n->{"net_exc"} = $n->lat_contri("exc");
	$n->{"net_inh"} = $n->lat_contri("inh");

	# New netinput: do not use g() -- for now
	$n->{"new_eta"} = $n->g( $n->{"gamma_aff"} * $n->{"net_aff"}
		+ $n->{"gamma_int"} * $n->{"net_int"}
		+ $n->{"gamma_exc"} * $n->{"net_exc"}
		- $n->{"gamma_inh"} * $n->{"net_inh"});

	$n->{"new_eta"} = (1.0-$n->{"noise"})*$n->{"new_eta"}
			+ $n->{"noise"}*rand(1.0);

	# New threshold
	$n->{"new_theta"} = $n->{"theta_base"} 
	      	+ $n->{"tau"}*($n->{"theta"}*$n->{"loop_decay"}+$n->{"spike"});

	if ($n->{"hard_refractory"}>0) {
		$n->{"new_theta"} += $n->Heav($n->{"r"})*$n->{"hard_refractory"};
	}

	# Membrane voltage
	$n->{"v"} = $n->{"eta"} 
		  - ($n->{"theta"} - $n->{"theta_base"})
		  + (($n->{"spike"}&&(!$n->{"novisible_spike"}))?1.0:0.0);
	# truncate membrande voltage if it goes below -2.0
	if ($n->{"v"}<-2.0) { $n->{"v"} = -2.0; } 
		
	# Refractory period
	if ($n->{"hard_refractory"}>0) {
	  $n->{"new_r"} =
		$n->{"r"} 
		+ $n->{"hard_refractory"}*$n->Heav($n->{"eta"} - $n->{"theta"})
		- 1;
	  $n->{"new_r"} *= $n->Heav($n->{"new_r"});
	  # if ($n->{"new_r"}<0) { $n->{"new_r"} = 0.0; }
	}
}

#----------------------------------------
#	Advance state : copy new values into current variables
#----------------------------------------
sub advance_state {

	my $n = shift;

	$n->{"s_exc"} = $n->{"new_s_exc"};
	$n->{"s_inh"} = $n->{"new_s_inh"};
	$n->{"s_int"} = $n->{"new_s_int"};
	$n->{"eta"}   = $n->{"new_eta"};
	$n->{"theta"} = $n->{"new_theta"};
	$n->{"r"} = $n->{"new_r"};

}

#----------------------------------------
#	Afferent contribution
#----------------------------------------
sub aff_contri {

	# Must fix this to have RFs and etc. Currently, it just provides
	# fixed input patterns.
	my $n = shift;
 	my $inp = $n->{"input_select"};
	my $x = $n->{"x"};
	my $netref = $n->{"network"};
	my $neurons = $#$netref;
   	my $input;

	$input = rand(1.0) if ($inp eq "random_pulse");
	$input =((sin($n->{"t"}*3.14/180.0)+1.0)/2.0) if ($inp eq "sine_pulse");
	$input = (($n->{"t"}/50)%2) if ($inp eq "square_pulse");
	$input = (1.0) if ($inp eq "constant");

	if ($inp eq "4objects") {
		my $div = $neurons/9.0;
		if ((int($x/$div))%2==0) {
			$input = 0;
		} else { 
			$input = 1;
		}
	} elsif ($inp eq "3objects") {
		my $div = $neurons/7.0;
		if ((int($x/$div))%2==0) {
			$input = 0;
		} else { 
			$input = 1;
		}
	} elsif ($inp eq "2objects") {
		my $div = $neurons/5.0;
		if ((int($x/$div))%2==0) {
			$input = 0;
		} else { 
			$input = 1;
		}
	} elsif ($inp eq "2objects-asymmetric") {
		my $div = $neurons/6.0;
		if ((int($x/$div))==1||(int($x/$div))==3||(int($x/$div))== 4){
			$input = 1;
		} else { 
			$input = 0;
		}
	} elsif ($inp eq "2objects-asymmetric2") {
		my $div = $neurons/6.0;
		if ((int($x/$div))==1||(int($x/$div))==3||(int($x/$div))== 4||(int($x/$div))==5){
			$input = 1;
		} else { 
			$input = 0;
		}
	} elsif ($inp eq "gaussian") {
		my $u = $neurons/2.0;
		my $std = $neurons/6.0;
		$input = (&gaussian($x, $u, $std));
	} elsif ($inp eq "short") {
		my $div = $neurons/3.0;
		if ($x>= $div && $x <= 2*$div) {
			$input = 1;
		} else {
			$input = 0;
		}
	}
	
	return ((1.0-$n->{"aff_noise"})*$input + $n->{"aff_noise"}*rand(1.0));

}

#----------------------------------------
#	Intercolumnar contribution
#----------------------------------------
sub int_contri {

	my $n = shift;
	my $sum = 0.0;
	return $sum;
	
}

#----------------------------------------
#	Lateral contribution
#----------------------------------------
sub lat_contri {

	my $n = shift;
  	my $type = shift;
	my $sum = 0.0;

	my $x = $n->{"x"};
	my $wref = $n->{$type."_w"};
	my $networkref = $n->{"network"};

	# print STDERR "[".join(",",keys(%$wref))."]\n";

	foreach $key (keys(%$wref)) {
		my $srcref = $networkref->[$key];
		$sum += $srcref->{"s_".$type} * $wref->{"$key"};
	}

	return $sum;

}

#----------------------------------------
#	show isi histogram
#----------------------------------------
sub show_isi_histogram {

	my $n = shift;
   	my $tag = shift;

	my $isi_hist = $n->{"isi_histogram"};
	foreach $isi (sort{$a <=> $b}(keys(%$isi_hist))) {
		print "$isi ".($isi_hist->{"$isi"})." $tag\n";
	}
}

#----------------------------------------
#	Add connection
#----------------------------------------
sub connect {

  	my ($n,$src_n,$connection_type,$weight,$wgt_profile,$radius,$center) = @_;

	# my $networkref = $n->{"network"};
	# my $src_num = $src_n->{"x"};

	if (!exists($n->{$connection_type."_w"})) {
		print STDERR "neuron->connect(): connection $connection_type is unknown\n";
		return;
	}
	
	my $wref = $n->{"$connection_type"."_w"};

	# 1. Uniform Weight
	if ($wgt_profile eq "uniform") { 
		$wref->{"$src_n"} = $weight ; return;
	}
	# 2. Gaussian: g(x,u,s)=1/(sqrt(2.0*3.141592)*s)*exp(-(x-u)**2/(2*s**2))
	elsif ($wgt_profile eq "gaussian"){ 
		my $sigma = $radius/3.0;
		# my $diff = abs($n->{"x"}-$src_n);
		my $diff = abs($center-$src_n);
		$wref->{"$src_n"} = $weight * &gaussian($diff,0,$sigma);
		print "$diff $wref->{$src_n} gaussian\n";
	}
	# 3. DoG -- diff of gaussian  only use with exc only connections
	elsif ($wgt_profile eq "DoG"){ 
		if ($n->{"inh_rad"}>0) {
			print STDERR "connect: don't use this with inh connections\n";
			exit;
		}
		my $sigma = $radius/6.0;
		my $diff = abs($n->{"x"}-$src_n);
		$wref->{"$src_n"} = $weight 
			* ( &gaussian($diff,0,$sigma) 
			    - &gaussian($diff,0,$sigma*2.0));
		print "$diff $wref->{$src_n} DoG\n";
	}
	# Default is uniform
	else { # defaut is uniform weight
		$wref->{"$src_n"} = $weight ; return;
	}
}

#----------------------------------------
# 	Connect within given range 
#----------------------------------------
sub connect_within_range {

	my ($n,$min,$max,$type,$weight,$recurse_flag,$wgt_prof,$rad,$center) = @_;

	my $i=0;

	for ($i=$min; $i<=$max; ++$i) {
		if ($recurse_flag==0 && $i==$n->{"x"}) {
			next;
		}
		$n->connect($i,$type,$weight,$wgt_prof,$rad,$center);
	}
}

#----------------------------------------
#	Normalize weights
#----------------------------------------
sub normalize_weights {

  	my ($n,$connection_type) = @_;

	if (!exists($n->{$connection_type."_w"})) {
		print STDERR "neuron->normalize(): connection $connection_type is unknown\n";
		return;
	}
	
	my $wref = $n->{"$connection_type"."_w"};

	my @keys = keys(%$wref);
	my $key;
  	my $sum = 0.0;

	# Get sum
	foreach $key (@keys) {
		$sum += $wref->{"$key"};
	}

	# Normalize
	if ($sum ==0) { return; }
	foreach $key (@keys) {
		$wref->{"$key"} /= $sum;
	}

}

#----------------------------------------
# Heaveside step function
#----------------------------------------
sub Heav {

  my ($n,$x) = @_;
  return (($x>0.0)?1.0:0.0);
  # return ((rand(10.0)<1.0)?1.0:0.0);
  # return 1;

}

#----------------------------------------
# Sigmoid
#----------------------------------------
sub g {

  my ($n,$x) = @_;
  if ($x < $n->{"alpha"}) {
	return 0.0;
  } elsif ($x < $n->{"beta"}) {
	return (($x-$n->{"alpha"})/($n->{"beta"}-$n->{"alpha"}));
  } else {
	return 1;
  }

}

#----------------------------------------
# plot_weight
#----------------------------------------
sub plot_weights {

  my ($n,$type,$offset) = @_;
  my $wref = $n->{"$type"."_w"};
  my @keys = sort{$a <=> $b}(keys(%$wref));
  my $key;
  foreach $key (@keys) {
	print "".($key+$offset)." ".($wref->{"$key"})." plotw\n";
  }

}

#----------------------------------------
# Gaussian
#----------------------------------------
sub gaussian {

  my ($x,$u,$s) = @_;	# u=mean, s=stdev

  return (1/(sqrt(2.0*3.141592)*$s)*exp(-($x-$u)**2/(2*$s**2)));

}


#----------------------------------------------------------------------------
# Network class
#----------------------------------------------------------------------------
package Network;

sub new {

	my $spec = shift;

 	my $network = {}; # reference to an anonymous hash

	# Copy attributes
	foreach $key (keys(%$spec)) {
		$network->{"$key"} = $spec->{"$key"};
	}
	
	# init weights
	$network->{"exc_w"} = {}; 
	$network->{"inh_w"} = {}; 
	$network->{"aff_w"} = {}; 
	$network->{"int_w"} = {}; 
	
	bless $network, 'Neuron';
	return $network; # return reference to anonymous hash

}

